__version__ = '0.3.0'

from .gselu import ElectrodePositionsModel

#import gselu
#from gselu import ElectrodePositionsModel
#
#import sys
#sys.modules['gselu'] = gselu
#sys.modules['gselu.gselu'] = gselu
#sys.modules['gselu.ElectrodePositionsModel'] = ElectrodePositionsModel
#sys.modules['gselu.gselu.ElectrodePositionsModel'] = ElectrodePositionsModel
#sys.modules['ElectrodePositionsModel'] = ElectrodePositionsModel
